# -*- coding: utf-8 -*-
from urlquote._native import ffi, lib
from .quoting import DEFAULT_QUOTING
import six
import threading

threadlocal = threading.local()


def _ensure_buffer():
    # This buffer is passed to the C interface in order to obtain the quoted string. It will be
    # reallocated automatically by `_native_quote` and `_native_unquote` if its size should not be
    # large enough. It is ok to reset this buffer to a smaller value but it always needs to be a valid
    # buffer.
    # To prevent race conditions on the buffer between threads, the buffer is made threadlocal
    if not hasattr(threadlocal, "buffer"):
        threadlocal.buffer = ffi.new("uint8_t[]", 1)


def _native_quote(value, quoting):
    """
    Urlencodes the given bytes
    """
    _ensure_buffer()
    buffer = threadlocal.buffer
    buffer_len = len(buffer)
    quoted_len = lib.quote(value, len(value), buffer, buffer_len, quoting)
    if quoted_len > buffer_len:
        # Our buffer has not been big enough to hold the quoted url. Let's allocate a buffer large
        # enough and try again.
        buffer = ffi.new("uint8_t[]", quoted_len)
        lib.quote(value, len(value), buffer, quoted_len, quoting)

    return ffi.string(buffer, quoted_len)


def _native_unquote(value):
    """
    Urldecodes the given bytes
    """
    _ensure_buffer()
    buffer = threadlocal.buffer
    buffer_len = len(buffer)
    unquoted_len = lib.unquote(value, len(value), buffer, buffer_len)
    if unquoted_len > buffer_len:
        # Our buffer has not been big enough to hold the unquoted url. Let's allocate a buffer large
        # enough and try again.
        buffer = ffi.new("uint8_t[]", unquoted_len)
        lib.unquote(value, len(value), buffer, unquoted_len)

    return ffi.string(buffer, unquoted_len)


def quote(value, quoting=DEFAULT_QUOTING):
    """
    Performs string encoding and urlencodes the given string. Always returns utf-8 encoded bytes.
    """
    if not isinstance(value, six.binary_type):
        if not isinstance(value, six.text_type):
            value = str(value)
        value = value.encode("utf-8")

    return _native_quote(value, quoting)


def unquote(value):
    """
    Decodes a urlencoded string and performs necessary decoding depending on the used Python version.
    """
    if not isinstance(value, six.binary_type):
        if not isinstance(value, six.text_type):
            value = str(value)
        value = value.encode("utf-8")

    return _native_unquote(value)



from urlquote._native import lib

# All ASCII charcters less than hexidecimal 20 and greater than 7E are encoded.  This includes
# special charcters such as line feed, carriage return, NULL, etc.
SIMPLE_QUOTING = lib.SIMPLE_QUOTING

# This quoting is used for url path components.
#
# Aside from special chacters defined in the `SIMPLE_QUOTING`, space, double quote ("), hash (#),
# inequality qualifiers (<), (>), backtick (`), question mark (?), and curly brackets ({), (}) are
# encoded.
DEFAULT_QUOTING = lib.DEFAULT_QUOTING

# This quoting is used for url query strings.
#
# Aside from special chacters defined in the `SIMPLE_QUOTING`, space, double quote ("), hash (#),
# and inequality qualifiers (<), (>) are encoded.
QUERY_QUOTING = lib.QUERY_QUOTING

# This quoting is used for on '/'-separated path segment
#
# Aside from special chacters defined in the `SIMPLE_QUOTING`, space, double quote ("), hash (#),
# inequality qualifiers (<), (>), backtick (`), question mark (?), and curly brackets ({), (}),
# percent sign (%), forward slash (/) are encoded.
PATH_SEGMENT_QUOTING = lib.PATH_SEGMENT_QUOTING

# This quoting is used for username and password.
#
# Aside from special chacters defined in the `SIMPLE_QUOTING`, space, double quote ("), hash (#),
# inequality qualifiers (<), (>), backtick (`), question mark (?), and curly brackets ({), (}),
# forward slash (/), colon (:), semi-colon (;), equality (=), at (@), backslash (\\), square
# brackets ([), (]), caret (\^), and pipe (|) are encoded.
USERINFO_QUOTING = lib.USERINFO_QUOTING

# This is the default quoting used by urlib.parse.quote
#
# Quotes everything but alphanumeric letters, numbers and dash (-), underscore (_), slash (/),
# point (.) and tilde (~).
PYTHON_3_7_QUOTING = lib.PYTHON_3_7_QUOTING